/*
** $Id: main.c,v 1.13 2013/03/16 19:16:55 lpitcher Exp lpitcher $
**
** Purpose:	This program manages a semaphore file to track the number
**		of outstanding scripts executing. When the program detects
**		the start of the first execution of the target script, it
**		will, prior to executing the target script, run one or more
**		resource-creation scripts. When the program detects the end
**		of the final execution of the target script, it will, after
**		the execution of the target script, run one or more
**		resource-destruction scripts.
**
**		This behaviour permits automatic external creation of
**		required resources before the first target script executes,
**		and automatic cleanup of those resources after the last
**		target script completes.
**
**		The intended use of this program is as a hashbang shell
**		interpreter embedded directly in various scripts. The
**		resource creation and cleanup scripts are invoked to (for
**		example) mount and umount external filesystems, pre and
**		post process various data files, etc.
**
** Usage:	<command> [flags] resourcelist [/path/to/interpreter] /path/to/target
**
**		where	flags consists of
**			  -L
**			for extended log identification
**		and	resourcelist consists of
**			  resource[,resourcelist]
*/

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <libgen.h>
#include <string.h>
#include <limits.h>

#include "config.h"
#include "elist.h"
#include "resources.h"
#include "process.h"
#include "exec_shell.h"

/*
** Maximum number of arguments that program will accept
*/
#define MAXARG 4

struct pgmargs
{
  char *resourcelist;
  char *interpreter;
  char *targetscript;
  unsigned optLogName : 1;	/* 1 to extend syslog ID to contain pgmname */
};

char *PgmName,		/* name of this binary, for error messages */
     *JobScript,	/* name of JOB's target script, for message logging */
     *JobName;		/* string used in syslog for rmonitor logging */

static int ParseArgs(struct pgmargs *, int, char **);
static int getargs(int, char **, int, char **);
static int doList(elist *);

/*
** main()	first level process:
**		- validate arguments and build list of all resources to be managed
**		- fork off second level (child) process to manage the resource list
**		- terminate with returncode based on success of argops.handling, child launch
*/
int main(int argc, char **argv)
{
  int rc = 0;
  char *JobBase;	/* name of JOB's base directory, for message logging */

  struct pgmargs options;

  PgmName = basename(strdup(argv[0]));

  if (ParseArgs(&options,argc,argv) == 0)
  {
    elist *targetsList;

    /* derive pgmname, launch directory name, script name for stderr & syslogging */
    JobBase   = basename(dirname(realpath(strdup(options.targetscript),NULL)));
    JobScript = basename(strdup(options.targetscript));

    if (options.optLogName)
    {
       JobName = malloc(strlen(PgmName) + 2 + strlen(JobScript) + 1);
       sprintf(JobName,"%s: %s",PgmName,JobScript);
    }
    else JobName   = JobScript;

    /* Parent logs with name of source directory */
    openlog(JobBase,LOG_CONS,SYSLOG_FACILITY);

    /* validate the resource names, build the list */
    if (targetsList = buildList(options.resourcelist,options.interpreter,options.targetscript))
    {
         pid_t worker;

         switch (worker = fork())
         {
           case 0: /* child */
             {
		openlog(JobName,LOG_CONS|LOG_PID,SYSLOG_FACILITY);
		syslog(SYSLOG_LEVEL_MESSAGE,"JOB Begins");

		rc = doList(targetsList);	/* returncode from target, or RUNERROR */
                if (rc == RUNERROR) rc = FAILED_EXEC;

		syslog(SYSLOG_LEVEL_MESSAGE,"JOB Status %d",rc);
             }
             break;

           case -1: /* parent, fork failed */
             syslog(SYSLOG_LEVEL_ERROR,"JOB %s Failed to launch",JobScript);
             fprintf(stderr,"JOB %s Failed to launch\n",JobScript);
             rc = FAILED_LAUNCH;
             break;

           default:
             syslog(SYSLOG_LEVEL_MESSAGE,"JOB %s[%lu] Launched",JobScript,(unsigned long)worker);
             break;
         }
    }
    else rc = FAILED_LAUNCH;
    closelog();
  }
  else rc = FAILED_ARGS;

  return rc;
}

/* ParseArgs()	extracts resourcelist, interpreter, and script
**		from the argv argument list given to main
** Accepts:	a struct parmargs, to store extracted arguments in
**		an int count of the number of commandline arguments, and
**		a pointer to a character pointer array, containing argument strings
** Updates:	a struct parmargs, with the extracted validated arguments
** Returns:	0 indicating success, or
**		1 indicating errors/failure
*/
static int ParseArgs(struct pgmargs *options, int argc, char **argv)
{
  char *arg[MAXARG], *flags;
  int narg, error = 0;

  options->resourcelist = options->interpreter = options->targetscript = NULL;
  options->optLogName = 0;

  switch (getargs(MAXARG, arg, argc, argv))
  {
    case 2:
      /*
      ** (command) resourcelist script
      */
      flags = NULL;
      options->resourcelist = arg[0];
      options->interpreter  = strdup(DEFAULT_SHELL);
      options->targetscript = arg[1];      
      break;

    case 3:
      /*
      ** (command) resourcelist interpreter script
      ** OR
      ** (command) flags resourcelist script
      */
      if (arg[0][0] == '-')
      { /* (command) flags resourcelist script */
        flags = arg[0];
        options->resourcelist = arg[1];
        options->interpreter  = strdup(DEFAULT_SHELL);
      }
      else
      { /* (command) resourcelist interpreter script */
        flags = NULL;
        options->resourcelist = arg[0];
        options->interpreter  = arg[1];
      }
      options->targetscript = arg[2];
      break;

    case 4:
      /* (command) flags resourcelist interpreter script */
      flags = arg[0];
      options->resourcelist = arg[1];
      options->interpreter  = arg[2];
      options->targetscript = arg[3];
      break;

    default:
      ++error;
      break;
  }

  if (flags && !error)
  {
    if (*flags == '-')
    {
      for (++flags; *flags; ++flags)
      {
        switch (*flags)
        {
          case 'L':
            options->optLogName = 1;
            break;

          default:
            fprintf(stderr, "%s: ERROR: Unrecognized flag '%c'\n", PgmName, *flags);
            ++error;
            break;
        }
      }
    }
    else
    {
      fprintf(stderr, "%s: ERROR: Expected flag, received \"%s\"\n", PgmName, flags);
      ++error;
    }
  }

  if (error)
    fprintf(stderr,
            "Usage: %s [-L] resourcename[,resourcename...] [interpreter] script\n",
            PgmName);

  return (error > 0);
}

/*
** getargs()	pull arguments out of the the argv array
**
** Accepts:	a count of the number of arguments to pull
**		a pointer to an array of character pointers,
**		  in which to place pulled arguments
**		a count of the number of arguments passed to main()
**		a pointer to an array of character pointers,
**		  in which the commandline arguments are stored
** Updates:	the array of character pointers in which the
**		  pulled arguments are placed
** Returns:	the number of arguments pulled, or 0 on error
**
** Note:	Hashbangs will pass in an argc == 3, with
**		  argv[0] set to the programname,
**		  argv[1] set to the remaining hashbanged arguments,
**		    strung together
**		  argv[2] set to the path to the hashbanged script
**		Because of this sort of invocation, we have to split out
**		the expected arguments from the argv[1] string.
**		Other invocation methods do not have this limitation/quirk.
*/
static int getargs(int maxarg, char **arg, int argc, char **argv)
{
  char *seps=" \t",	/* argument separator characters for strtok() */
       *p;		/* pointer to strtok()ed argument */
  int narg,		/* number of arguments, after argv[1] separation */
      argn;		/* argv counter */

  if (argc < 2)
  {
    fprintf(stderr,"%s: ERROR: Missing commandline arguments\n",PgmName);
    return 0;
  }

  /* split up argv[1], if necessary */
  for (narg = 0, p = strtok(strdup(argv[1]),seps);
       narg < 3 && p;
       ++narg, p = strtok(NULL,seps))
    arg[narg] = p;

  if (p)
  {
    fprintf(stderr,"%s: ERROR: Excess commandline arguments given\n",PgmName);
    return 0;
  }

  /* capture remaining arguments */
  for (argn = 2; argn < argc && narg < maxarg; ++argn)
    arg[narg++] = argv[argn];

  if (argn < argc)
  {
    fprintf(stderr,"%s: ERROR: Excess commandline arguments given\n",PgmName);
    return 0;
  }

  return narg;
}

/*
** doList()	launch required preprocessors, target script, and postprocessors
**
** Accepts:	a pointer to the top struct eList node in a list of nodes
**		describing the required resources and target script
**
** Updates:	semaphore files for each active resource
**		syslog with stdout, stderr from each resource processor
**		syslog with stdout, stderr from target script
**		closes semaphore files, updates related eList node
**
** Returns:	returncode from the target script, or RUNERROR
*/
static int doList(elist *this)
{
  int rc;

  if (this->ExecNext != NULL)	/* preprocessor/postprocessor entry */
  {
    if (doPrep(this->ExecTarget,this->ExecParm.ExecFile) == PROC_WORKED)
    {
      rc = doList(this->ExecNext);
      doPost(this->ExecTarget,this->ExecParm.ExecFile);
    }
    else rc = RUNERROR;
    close(this->ExecParm.ExecFile); this->ExecParm.ExecFile = -1;
  }
  else rc = doExec(this->ExecParm.ExecInterp,this->ExecTarget);

  return rc;
}


/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) February, 2013
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/
